#include <Arduino.h>
#include <NimBLEDevice.h>
#include <NimBLEHIDDevice.h>

#define controlPin 10

// ----- Global Objects -----
static NimBLEServer* pServer;
static NimBLEHIDDevice* pHIDDevice; 
static NimBLEService* pIMUService;
static NimBLECharacteristic* pActivateCameraCharacteristic;
static NimBLEAdvertising* pAdvertising;

static uint8_t hidReportDescriptor[] = {
  0x05, 0x0C,       // Usage Page (Consumer Devices)
  0x09, 0x01,       // Usage (Consumer Control)
  0xA1, 0x01,       // Collection (Application)
  0x85, 0x01,       //   Report ID (1)
  0x15, 0x00,       //   Logical Minimum (0)
  0x26, 0xFF, 0x03, //   Logical Maximum (1023)
  0x19, 0x00,       //   Usage Minimum (0)
  0x2A, 0xFF, 0x03, //   Usage Maximum (1023)
  0x75, 0x10,       //   Report Size (16)
  0x95, 0x01,       //   Report Count (1)
  0x81, 0x00,       //   Input (Data, Array)
  0xC0              // End Collection
};

// Variables will change:
int ledState = HIGH;        // the current state of the output pin
int controlPinState;            // the current reading from the input pin
int lastControlPinState = LOW;  // the previous reading from the input pin
bool initialized = false;

//  None of these are required as they will be handled by the library with defaults.
//                       Remove as you see fit for your needs                        
class ServerCallbacks : public NimBLEServerCallbacks {
    void onConnect(NimBLEServer* pServer, NimBLEConnInfo& connInfo) override {
        //Serial.printf("Client address: %s\n", connInfo.getAddress().toString().c_str());

        //
        //  We can use the connection handle here to ask for different connection parameters.
        //  Args: connection handle, min connection interval, max connection interval
        //  latency, supervision timeout.
        //  Units; Min/Max Intervals: 1.25 millisecond increments.
        //  Latency: number of intervals allowed to skip.
        //  Timeout: 10 millisecond increments.
        //
        pServer->updateConnParams(connInfo.getConnHandle(), 24, 48, 0, 180);
    }

    void onDisconnect(NimBLEServer* pServer, NimBLEConnInfo& connInfo, int reason) override {
        //Serial.printf("Client disconnected - start advertising\n");
        NimBLEDevice::startAdvertising();
    }

    void onMTUChange(uint16_t MTU, NimBLEConnInfo& connInfo) override {
        //Serial.printf("MTU updated: %u for connection ID: %u\n", MTU, connInfo.getConnHandle());
    }

    /********************* Security handled here *********************/
    uint32_t onPassKeyDisplay() override {
        //Serial.printf("Server Passkey Display\n");
        //
        // This should return a random 6 digit number for security
        //  or make your own static passkey as done here.
        //
        return 123456;
    }

    void onConfirmPassKey(NimBLEConnInfo& connInfo, uint32_t pass_key) override {
        //Serial.printf("The passkey YES/NO number: %" PRIu32 "\n", pass_key);
        // Inject false if passkeys don't match.
        NimBLEDevice::injectConfirmPasskey(connInfo, true);
    }

    void onAuthenticationComplete(NimBLEConnInfo& connInfo) override {
        // Check that encryption was successful, if not we disconnect the client
        if (!connInfo.isEncrypted()) {
            NimBLEDevice::getServer()->disconnect(connInfo.getConnHandle());
            //Serial.printf("Encrypt connection failed - disconnecting client\n");
            return;
        }

        //Serial.printf("Secured connection to: %s\n", connInfo.getAddress().toString().c_str());
    }
} serverCallbacks;

void initialize() {
  // Initialize NimBLE and set the device name
  NimBLEDevice::init("Nano ESP32 NimBLE Server");

  //
  // Set the IO capabilities of the device, each option will trigger a different pairing method.
  //  BLE_HS_IO_DISPLAY_ONLY    - Passkey pairing
  //  BLE_HS_IO_DISPLAY_YESNO   - Numeric comparison pairing
  //  BLE_HS_IO_NO_INPUT_OUTPUT - DEFAULT setting - just works pairing
  //
  // NimBLEDevice::setSecurityIOCap(BLE_HS_IO_DISPLAY_ONLY); // use passkey
  // NimBLEDevice::setSecurityIOCap(BLE_HS_IO_DISPLAY_YESNO); //use numeric comparison

  //
  //  2 different ways to set security - both calls achieve the same result.
  //  no bonding, no man in the middle protection, BLE secure connections.
  //
  //  These are the default values, only shown here for demonstration.
  //
    NimBLEDevice::setSecurityAuth(true, false, true);

  //NimBLEDevice::setSecurityAuth(BLE_SM_PAIR_AUTHREQ_BOND | BLE_SM_PAIR_AUTHREQ_MITM | BLE_SM_PAIR_AUTHREQ_SC);
  pServer = NimBLEDevice::createServer();
  pServer->setCallbacks(&serverCallbacks);


  // ----- Initialize the HID device -----
  pHIDDevice = new NimBLEHIDDevice(pServer);
  pHIDDevice->setManufacturer("Nano ESP32");
  // PnP information: Vendor ID Source, Vendor ID, Product ID, Version.
  pHIDDevice->setPnp(0x02, 0x1234, 0x5678, 0x0100);
  pHIDDevice->setHidInfo(0x00, 0x01);
  
  // Set the report map (cast away const for compatibility)
  pHIDDevice->setReportMap((uint8_t*)hidReportDescriptor, sizeof(hidReportDescriptor));
  
  // Create an input report for Consumer Control (Report ID 1)
  pActivateCameraCharacteristic = pHIDDevice->getInputReport(1);
  
  // Start HID services
  pHIDDevice->startServices();

  // ----- Start Advertising -----
  NimBLEAdvertising* pAdvertising = NimBLEDevice::getAdvertising();
  // Set an appearance (using HID_KEYBOARD for compatibility; adjust if needed)
  pAdvertising->setName("Nano ESP32 Camera Control");
  pAdvertising->setAppearance(HID_KEYBOARD);
  pAdvertising->addServiceUUID(pHIDDevice->getHidService()->getUUID());
  pAdvertising->start();

  //NimBLEDevice::startAdvertising();
  //Serial.println("Advertising started. Waiting for connection...");
}

void activateCamera() {
  // 0xE9 is the HID usage code for Volume Up.
  uint16_t record = 0x00E9;
  //Serial.println("Sending Record command...");
  
  // Send key press (consumer control report)
  pActivateCameraCharacteristic->setValue((uint8_t*)&record, sizeof(record));
  pActivateCameraCharacteristic->notify();
  delay(10);
  
  // Send key release (report value 0)
  record = 0;
  pActivateCameraCharacteristic->setValue((uint8_t*)&record, sizeof(record));
  pActivateCameraCharacteristic->notify();
}

void setup() {
  //Serial.begin(115200);
  //while (!Serial) delay(10);

  pinMode(controlPin, INPUT_PULLUP);

  initialize();
}

void loop() {
  // read the state of the switch into a local variable:
  controlPinState = digitalRead(controlPin);

  if (!controlPinState && lastControlPinState) {
    initialized = true;
    //Serial.println("button pressed");
  }
  if (controlPinState && !lastControlPinState && initialized) {
    //Serial.println("button released");
    activateCamera();
  } 
  lastControlPinState = controlPinState;
}
